--===================================================================--
-- kNavi.lua (creates navigation for thumbnails)
--
-- by Kwiksher
-- version 1: April 12, 2012
--===================================================================--
-- Sintaxe:
-- Navigation.show( g1, { backColor = {255,255,0}, alpha = alpha, totPages = numPages, curPage = curPage, thumbW = 120, thumbH = 70, imageDir = imgDir, dire = "right"} )
-- navGroup holds all navigation elements
-- navBack holds the rect or image acting as background for the navigation
-- navItems holds all thumbnails

--[[IN THE PAGE:
       local show = 0; local knavi = 1
       local function g1Drag (event )
          if knavi == 0 then
          	print("open navigation")
          	Navigation.new( "teste", { backColor = {255,255,0}, totPages = numPages, curPage = curPage, thumbW = 120, thumbH = 70, imageDir = imgDir, dire = "right"} )
          	knavi = 1
          end
          if show == 0 then
          	Navigation.show()
          	show = 1
          else
          	Navigation.hide()
          	show = 0
          end
       end
       g1:addEventListener( "touch", g1Drag )
--]]

local Navigation = {}
local navShow = 0

Navigation.new = function( obj, params )
--print("building navigation")
--print( "Navigation.show", obj._name, event )
	params = params or {}

	local audios = params.audio

	local _W = display.contentWidth
	local _H = display.contentHeight


	--creates group for the full navigation
	navGroup = display.newGroup()
	navGroup.x = 0; navGroup.y = 0

	local pageList = {}

	--Loads all images to the table
	for i = 1,params.totPages do
		pageList [i] = params.imageDir.."p"..i.."_thumb.jpg"
    end

    --sets background
    local spacer = 10 --pixels distance between images
    local groupW
    local groupH
    if (params.dire == "bottom" or params.dire == "top") then
    	groupW = (spacer * (params.totPages + 1)) + (params.totPages * params.thumbW)
    	groupH = (spacer * 2) + params.thumbH
    	if (groupW < _W ) then groupW = _W end
	else
		groupH = (spacer * (params.totPages + 1)) + (params.totPages * params.thumbH)
		groupW = (spacer * 2) + params.thumbW
    	if (groupH < _H ) then groupH = _H end
	end


    local navBack
    --Add code in the future to accept params.backFile
    if (params.dire == "bottom") then
    	 navBack= display.newRect(0, _H - (params.thumbH + (spacer * 2)), _W, _H - (params.thumbH + (spacer * 2)))
    	 navBack:setFillColor(params.backColor[1],params.backColor[2],params.backColor[3])
    elseif (params.dire == "top") then
    	 navBack= display.newRect(0, 0, _W, (params.thumbH + (spacer * 2)))
    	 navBack:setFillColor(params.backColor[1],params.backColor[2],params.backColor[3])
    elseif (params.dire == "left") then
    	 navBack= display.newRect(0, 0, groupW, groupH)
    	 navBack:setFillColor(params.backColor[1],params.backColor[2],params.backColor[3])
	elseif (params.dire == "right") then
    	 navBack= display.newRect(_W - groupW, 0, groupW, groupH)
    	 navBack:setFillColor(params.backColor[1],params.backColor[2],params.backColor[3])
    end

    navBack.alpha = params.alpha
    navGroup:insert(navBack)

    local navItems = display.newGroup()
	navGroup:insert(navItems)

	--controls thumbnails actions
	local menuItemTap = function ( event )
		if curPage  ~= event.target.id then -- director.getCurrBookPageNum() ~= event.target.id then
			local togo = "page_"..event.target.id

			if (#audios > 0) then
				--dispose audios
				for x = 1,#audios do
					audio.stop(audios[x][1]); audio.dispose(audios[x][2]); audios[x][2] = nil
				end
			end

			director:changeScene ( imgDir..togo, true )
		end
	end
    --builds thumbnails
      	local menuTab = {}
      	for i=1, table.maxn( pageList ) do

		menuTab[i] = display.newImage( pageList[i])
		menuTab[i]:setReferencePoint( display.TopLeftReferencePoint )
		menuTab[i].id = i
		menuTab[i]:addEventListener( "tap", menuItemTap )

		-- navigation positioning
		if i == 1 and params.dire == "bottom" then
			menuTab[i].x = spacer
			menuTab[i].y = _H - (spacer + menuTab[i].height)
		elseif i == 1 and params.dire == "top" then
			menuTab[i].x = spacer
			menuTab[i].y = 0 + spacer
		elseif i == 1 and params.dire == "left" then
			menuTab[i].x = spacer
			menuTab[i].y = 0 + spacer
		elseif i == 1 and params.dire == "right" then
			menuTab[i].x = _W - (spacer + menuTab[i].width)
			menuTab[i].y = 0 + spacer
		elseif i ~= 1 and params.dire == "bottom" or params.dire == "top" then
			menuTab[i].x = menuTab[i-1].x + spacer + menuTab[i-1].width
			menuTab[i].y = menuTab[i-1].y
		elseif i ~= 1 and params.dire == "left" or params.dire == "right" then
			menuTab[i].x = menuTab[i-1].x
			menuTab[i].y = menuTab[i-1].y + spacer + menuTab[i].height
		end

		navItems:insert( menuTab[i] )

	end

	--add drag
	local ang
	if params.dire == "left" or params.dire =="right" then
		ang = 90
	elseif params.dire == "bottom" or params.dire =="top" then
		ang = 0
	end

    --sets navigation out of screen
    --UNCOMMENT THIS LINE WHEN BUTTONS ARE ACTIVE
   -- navGroup.x = -2500


       local function naviDrag (event )
          if event.phase == "moved" then

          	if (params.dire == "bottom" or params.dire == "top") then
          		if (navItems.x > 0) then navItems.x = 0 end
          		if (navItems.x < ((navItems.width - _W)*-1)) then navItems.x = ((navItems.width + (spacer*2) - _W)*-1) end
          	else
          		if (navItems.y > 0) then navItems.y = 0 end
          		if (navItems.y < ((navItems.height - _H)*-1)) then navItems.y = ((navItems.height + (spacer*2) - _H)*-1) end
          	end

          elseif event.phase == "ended" then
          end
       end

	   -- Only activates drag if number of thumb pages are bigger than the page
       if (navItems.width > _W or navItems.height > _H) then
       	   MultiTouch.activate( navItems, "move", "single", {constrainAngle = ang}) --, xBounds = {xbi, xbe}} )
           navItems:addEventListener( MultiTouch.MULTITOUCH_EVENT, naviDrag )
    	end


     Navigation.hide = function( obj, params )
		--print ("hide navigator")
		navGroup.x = -2500
	end

	Navigation.show = function( obj, params )
		--print ("show navigator")
		navGroup.x = 0
	end

	return obj

end







return Navigation
